class MissingRadical(UserWarning):
    """ Error thrown on missing radical """


class UnknownModeleConfigurationKey(UserWarning):
    """ Error thrown when a configuration key is absolutely unknown """
